import torch
from typing import Any, Dict, List, NamedTuple, Tuple, Union
from espnet.nets.beam_search import BeamSearch, Hypothesis
from espnet.nets.batch_beam_search import BatchBeamSearch
from espnet.nets.e2e_asr_common import end_detect

import logging

logger = logging.getLogger("dolphin")


class SimpleBeamSearch(BatchBeamSearch):
    """Simple beam search decoder.
    Just when the hyps equals to beam size, it will stop decoding."""
        
    def forward(
        self,
        x: torch.Tensor,
        maxlenratio: float = 0.0,
        minlenratio: float = 0.0,
        pre_x: torch.Tensor = None,
    ) -> List[Hypothesis]:
        """Perform beam search.

        Args:
            x (torch.Tensor): Encoded speech feature (T, D)
            maxlenratio (float): Input length ratio to obtain max output length.
                If maxlenratio=0.0 (default), it uses a end-detect function
                to automatically find maximum hypothesis lengths
                If maxlenratio<0.0, its absolute value is interpreted
                as a constant max output length.
            minlenratio (float): Input length ratio to obtain min output length.
                If minlenratio<0.0, its absolute value is interpreted
                as a constant min output length.
            pre_x (torch.Tensor): Encoded speech feature for sequential attn (T, D)
                Sequential attn computes attn first on pre_x then on x,
                thereby attending to two sources in sequence.

        Returns:
            list[Hypothesis]: N-best decoding results

        """
        logger.debug("Simple beam search decoding...")
        # set length bounds
        if pre_x is not None:
            inp = pre_x
        else:
            inp = x
        if maxlenratio == 0:
            maxlen = inp.shape[0]
        elif maxlenratio < 0:
            maxlen = -1 * int(maxlenratio)
        else:
            maxlen = max(1, int(maxlenratio * inp.size(0)))

        if minlenratio < 0:
            minlen = -1 * int(minlenratio)
        else:
            minlen = int(minlenratio * inp.size(0))
        logger.debug("decoder input length: " + str(inp.shape[0]))
        logger.debug("max output length: " + str(maxlen))
        logger.debug("min output length: " + str(minlen))

        # main loop of prefix search
        running_hyps = self.init_hyp(x if pre_x is None else pre_x)
        ended_hyps = []
        for i in range(maxlen):
            logger.debug("position " + str(i))
            best = self.search(running_hyps, x, pre_x=pre_x)
            # post process of one iteration
            running_hyps = self.post_process(
                i, maxlen, minlen, maxlenratio, best, ended_hyps
            )
            # simple beam search stop condition
            if len(ended_hyps) >= self.beam_size:
                logger.debug(f"end detected for beam sizs: {self.beam_size}")
                break 
            # end detection
            if maxlenratio == 0.0 and end_detect([h.asdict() for h in ended_hyps], i):
                logger.debug(f"end detected at {i}")
                break
            if len(running_hyps) == 0:
                logger.debug("no hypothesis. Finish decoding.")
                break
            else:
                logger.debug(f"remained hypotheses: {len(running_hyps)}")

        if self.normalize_length:
            # Note (Jinchuan): -1 since hyp starts with <sos> and
            # initially has score of 0.0
            nbest_hyps = sorted(
                ended_hyps, key=lambda x: x.score / (len(x.yseq) - 1), reverse=True
            )
        else:
            nbest_hyps = sorted(ended_hyps, key=lambda x: x.score, reverse=True)

        # check the number of hypotheses reaching to eos
        if len(nbest_hyps) == 0:
            logger.warning(
                "there is no N-best results, perform recognition "
                "again with smaller minlenratio."
            )
            return (
                []
                if minlenratio < 0.1
                else self.forward(x, maxlenratio, max(0.0, minlenratio - 0.1))
            )

        # report the best result
        best = nbest_hyps[0]
        for k, v in best.scores.items():
            logger.debug(
                f"{v:6.2f} * {self.weights[k]:3} = {v * self.weights[k]:6.2f} for {k}"
            )
        logger.debug(f"total log probability: {best.score:.2f}")
        logger.debug(f"normalized log probability: {best.score / len(best.yseq):.2f}")
        logger.debug(f"total number of ended hypotheses: {len(nbest_hyps)}")
        if self.token_list is not None:
            logger.debug(
                "best hypo: "
                + "".join([self.token_list[x] for x in best.yseq[1:-1]])
                + "\n"
            )
        if best.yseq[1:-1].shape[0] == maxlen:
            logger.warning(
                "best hypo length: {} == max output length: {}".format(
                    best.yseq[1:-1].shape[0], maxlen
                )
            )
            logger.warning(
                "decoding may be stopped by the max output length limitation, "
                + "please consider to increase the maxlenratio."
            )
        return nbest_hyps



import torch
import numpy as np
from typing import Any, Tuple

from espnet2.bin.s2t_inference import ScoreFilter


class DolphinScoreFilter(ScoreFilter):
    """DolphinSpeech2Text score filter."""

    def score(
        self, y: torch.Tensor, state: Any, x: torch.Tensor
    ) -> Tuple[torch.Tensor, Any]:
        """Score new token (required).

        Args:
            y (torch.Tensor): 1D torch.int64 prefix tokens.
            state: Scorer state for prefix tokens
            x (torch.Tensor): The encoder feature that generates ys.

        Returns:
            tuple[torch.Tensor, Any]: Tuple of
                scores for next token that has a shape of `(n_vocab)`
                and next state for ys

        """

        score = torch.zeros(
            self.vocab_size, dtype=self.param.dtype, device=self.param.device
        )
        
        if self.notimestamps in y:
            # Suppress timestamp tokens if we don't predict time
            score[self.first_time : self.last_time + 1] = -np.inf
        elif y[-4] == self.sos: # <sos> <lang> <region> <asr>
            # The first token must be a timestamp if we predict time
            score[: self.first_time] = -np.inf
            score[self.last_time + 1 :] = -np.inf
        else:
            prev_times = y[torch.logical_and(y >= self.first_time, y <= self.last_time)]
            if len(prev_times) % 2 == 1:
                # there are an odd number of timestamps, so the sentence is incomplete
                score[self.eos] = -np.inf
                # timestamps are monotonic
                score[self.first_time : prev_times[-1] + 1] = -np.inf
            else:
                # there are an even number of timestamps (all are paired)
                if y[-1] >= self.first_time and y[-1] <= self.last_time:
                    # the next tokon should be a timestamp or eos
                    score[: y[-1]] = -np.inf
                    score[self.last_time + 1 :] = -np.inf
                    score[self.eos] = 0.0
                else:
                    # this is an illegal hyp
                    score[:] = -np.inf

        return score, None



# encoding: utf8

SAMPLE_RATE = 16000
SPEECH_LENGTH = 30

NOTIME_SYMBOL = "<notimestamp>"
FIRST_TIME_SYMBOL = "<0.00>"
LAST_TIME_SYMBOL = "<30.00>"
FIRST_LANG_SYMBOL = "<ab>"
LAST_LANG_SYMBOL = "<zu>"
FIRST_REGION_SYMBOL = "<AD>"
LAST_REGION_SYMBOL = "<ZA>"



# encoding: utf8

import logging
import warnings

LOGGING_FORMAT="[%(asctime)s] [%(levelname)s] [%(filename)s:%(lineno)d:%(funcName)s] %(message)s"
logging.basicConfig(level=logging.INFO, format=LOGGING_FORMAT)
warnings.filterwarnings("ignore", category=FutureWarning)
# filter framework interanl logs
logging.getLogger("espnet").setLevel(logging.ERROR)
logging.getLogger("root").setLevel(logging.ERROR)
logging.getLogger("dolphin").setLevel(logging.INFO)

import yaml
import tqdm
import pydub
import logging
import hashlib
import os.path
import argparse
import dataclasses
import numpy as np
from pathlib import Path
from argparse import Namespace
from os.path import dirname, join, abspath, join
from distutils.util import strtobool
from typing import Union, Optional, Tuple, List

import torch
import modelscope
from modelscope.models.audio.funasr.model import GenericFunASR

from .audio import load_audio, convert_audio
from .model import DolphinSpeech2Text, TranscribeResult, TranscribeSegmentResult
from .languages import LANGUAGE_REGION_CODES, LANGUAGE_CODES
from .constants import SPEECH_LENGTH

logger = logging.getLogger("dolphin")

VAD_MODEL = "iic/speech_fsmn_vad_zh-cn-16k-common-pytorch"


MODELS = {
    "base": {
        "model_id": "DataoceanAI/dolphin-base",
        "download_url": "http://so-algorithm-prod.oss-cn-beijing.aliyuncs.com/models/dolphin/base.pt",
        "sha256": "688f0cdb26da2684a4eec200a432091920287585e8e332507cbe9c1ab6d77401",
        "config": {
            "encoder": {
                "output_size": 512,
                "attention_heads": 8,
                "cgmlp_linear_units": 2048,
                "num_blocks": 6,
                "linear_units": 2048,
            },
            "decoder": {
                "attention_heads": 8,
                "linear_units": 2048,
                "num_blocks": 6,
            }
        }
    },
    "small": {
        "model_id": "DataoceanAI/dolphin-small",
        "download_url": "http://so-algorithm-prod.oss-cn-beijing.aliyuncs.com/models/dolphin/small.pt",
        "sha256": "e5a52b9a713d294d5a2d929f5e7f6a18d951a8155ede80f935a74b76b0432b17",
        "config": {
            "encoder": {
                "output_size": 768,
                "attention_heads": 12,
                "cgmlp_linear_units": 3072,
                "num_blocks": 12,
                "linear_units": 1536,
            },
            "decoder": {
                "attention_heads": 12,
                "linear_units": 3072,
                "num_blocks": 12,
            }
        }
    },
}


def str2bool(value: str) -> bool:

    return bool(strtobool(value))


def parser_args() -> Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("audio", type=str, help="audio file path")
    parser.add_argument("--model", type=str, default="small", help="model name (default: small)")
    parser.add_argument("--model_dir", type=Path, default=None, help="model checkpoint download diretory")
    parser.add_argument("--lang_sym", type=str, default=None, help="language symbol (e.g. zh)")
    parser.add_argument("--region_sym", type=str, default=None, help="regiion symbol (e.g. CN)")
    parser.add_argument("--device", type=str, default=None, help="torch device (default: None)")
    parser.add_argument("--normalize_length", type=str2bool, default=False, help="whether to normalize length (default: false)")
    parser.add_argument("--padding_speech", type=str2bool, default=False, help="whether padding speech to 30 seconds (default: false)")
    parser.add_argument("--predict_time", type=str2bool, default=True, help="whether predict timestamp (default: true)")
    parser.add_argument("--beam_size", type=int, default=5, help="number of beams in beam search (default: 5)")
    parser.add_argument("--maxlenratio", type=float, default=0.0, help="Input length ratio to obtain max output length (default: 0.0)")
    parser.add_argument("--disable_model_hash_check", action="store_true", help="Disable model checkpoint hash check")

    args = parser.parse_args()
    return args


def detect_device():
    if torch.cuda.is_available():
        device = "cuda"
    elif torch.backends.mps.is_available() and torch.backends.mps.is_built():
        device = "mps"
    else:
        device = "cpu"

    return device


def seconds_to_hms(total_seconds: int):
    total_seconds = int(total_seconds)
    hours = total_seconds // 3600
    remaining_seconds = total_seconds % 3600
    minutes = remaining_seconds // 60
    seconds = remaining_seconds % 60
    if hours > 0:
        return f"{hours:02d}:{minutes:02d}:{seconds:02d}"
    else:
        return f"{minutes:02d}:{seconds:02d}"


def load_model(
    model_name: str,
    model_dir: Union[Path, str],
    device: Optional[Union[str, torch.device]] = None,
    **kwargs
) -> DolphinSpeech2Text:
    """
    Load DolphinSpeech2Text model.

    Args:
        model_name: model name (e.g. small)
        model_dir: model download directory
        device: the pytorch device

    Returns:
        DolphinSpeech2Text instance
    """
    if device is None:
        device = detect_device()
        logger.info(f"auto detect device: {device}")

    model_config = MODELS[model_name]["config"]
    train_cfg_file = join(dirname(abspath(__file__)), "assets/config.yaml")
    with open(train_cfg_file, "r", encoding="utf-8") as f:
        train_cfg = yaml.safe_load(f)
        train_cfg["encoder_conf"].update(**model_config["encoder"])
        train_cfg["decoder_conf"].update(**model_config["decoder"])

    if isinstance(model_dir, str):
        model_dir = Path(model_dir)

    model_file = model_dir / f"{model_name}.pt"
    download_model = True
    disable_model_hash_check = kwargs.get("disable_model_hash_check", False)
    if model_file.exists() and not disable_model_hash_check:
        with open(model_file, "rb") as f:
            model_bytes = f.read()
        if hashlib.sha256(model_bytes).hexdigest() == MODELS[model_name]["sha256"]:
            download_model = False
        else:
            model_file.unlink(missing_ok=True)
            logger.warning("model SHA256 chechsum mismatch, redownload model...")

    if download_model and not disable_model_hash_check:
        # Download model
        model_dir.mkdir(exist_ok=True)
        _download_from_modelscope(
            model_id=MODELS[model_name]["model_id"],
            local_dir=model_dir,
            allow_file_pattern=f"{model_name}.pt",
        )

    model = DolphinSpeech2Text(
        s2t_train_config=train_cfg,
        s2t_model_file=model_file,
        device=device,
        task_sym=kwargs.get("task_sym", "<asr>"),
        predict_time=kwargs.get("predict_time", True),
        **kwargs,
    )
    return model


def _download_from_modelscope(model_id: str, local_dir: str, allow_file_pattern: str):
    modelscope.snapshot_download(
        model_id=model_id,
        local_dir=local_dir,
        allow_file_pattern=allow_file_pattern,
        repo_type="model",
    )


def validate_lang_region(lang_sym: str, region_sym: str):

    if all([lang_sym, region_sym]):
        if f"{lang_sym}-{region_sym}" not in LANGUAGE_REGION_CODES:
            raise Exception("Unsupport language or region!")
    elif lang_sym and region_sym is None:
        if lang_sym not in LANGUAGE_CODES:
            raise Exception("Unsupport language!")

    return True


def transcribe_long(
        model: DolphinSpeech2Text,
        audio: str,
        lang_sym: str = None,
        region_sym: str = None,
        predict_time: bool = True,
        padding_speech: bool = False,
        **kwargs,
    ) -> List[TranscribeSegmentResult]:
    """
    Transcribe audio to text.

    Args:
        model: model instance
        audio: audio path
        lang_sym: language symbol (e.g. zh)
        region_sym: regiion symbol (e.g. CN)
        predict_time: whether predict timestamp (default: true)
        padding_speech: whether padding speech to 30 seconds (default: false)

    Returns:
        List[TranscribeSegmentResult]
    """
    results = []

    validate_lang_region(lang_sym, region_sym)

    logging.info("download vad model")
    vad_model_dir = Path(os.path.expanduser("~/.cache/dolphin/speech_fsmn_vad"))
    vad_model_dir.mkdir(exist_ok=True)
    _download_from_modelscope(VAD_MODEL, vad_model_dir, None)

    logger.info("loading vad model")
    vad_model = GenericFunASR(vad_model_dir, max_single_segment_time=SPEECH_LENGTH*1000, device="cpu")

    # convert audio to sample rate 16k Mono channel audio
    tmp_audio = f"{audio}.wav"
    convert_audio(audio, tmp_audio)

    logger.info("run vad model")
    segments = vad_model(input=tmp_audio, disable_pbar=True)[0]["value"]

    logger.info("decoding...")
    audio_segment = pydub.AudioSegment.from_wav(tmp_audio)
    for seg in segments:
        s, e = seg
        raw_data = audio_segment[s:e].raw_data
        waveform = np.frombuffer(raw_data, np.int16).flatten().astype(np.float32) / 32768.0
        result = model(
            speech=waveform,
            lang_sym=lang_sym,
            region_sym=region_sym,
            predict_time=predict_time,
            padding_speech=padding_speech,
        )

        st = seconds_to_hms(s/1000)
        et = seconds_to_hms(e/1000)
        logger.info(f"segment: {st} - {et}, lang: {result.language}, region: {result.region}, text: {result.text_nospecial}")
        result_json = dataclasses.asdict(result)
        result_json.update({
            "start": round(s/1000, 2),
            "end": round(e/1000, 2)
        })
        segment_result = TranscribeSegmentResult(**result_json)
        results.append(segment_result)

    # clean tmp audio file
    Path(tmp_audio).unlink(missing_ok=True)

    return results


def transcribe(
        model: DolphinSpeech2Text,
        audio: str,
        lang_sym: str = None,
        region_sym: str = None,
        predict_time: bool = True,
        padding_speech: bool = False,
        **kwargs,
    ) -> TranscribeResult:
    """
    Transcribe audio to text.

    Args:
        model: model instance
        audio: audio path
        lang_sym: language symbol (e.g. zh)
        region_sym: regiion symbol (e.g. CN)
        predict_time: whether predict timestamp (default: true)
        padding_speech: whether padding speech to 30 seconds (default: false)

    Returns:
        TranscribeResult
    """
    waveform = load_audio(audio)
    validate_lang_region(lang_sym, region_sym)

    logger.info("decoding...")
    result = model(
        speech=waveform,
        lang_sym=lang_sym,
        region_sym=region_sym,
        predict_time=predict_time,
        padding_speech=padding_speech
    )

    logger.info(f"decode result, rtf: {result.rtf}, language: {result.language}, region: {result.region}, text: {result.text}")
    return result


def cli():
    args = parser_args()

    model = args.model
    if model not in MODELS:
        logging.error(f"Unknown model {model}, Dolphin open source base, small model, please config the correct model.")
        return

    model_dir = args.model_dir if args.model_dir else os.path.expanduser("~/.cache/dolphin")
    model_dir = Path(model_dir)

    logger.info("loading asr model")
    device = args.device if args.device else detect_device()
    model_kwargs = {
        "device": device,
        "normalize_length": args.normalize_length,
        "beam_size": args.beam_size,
        "maxlenratio": args.maxlenratio,
        "disable_model_hash_check": args.disable_model_hash_check,
    }
    model_instance = load_model(model, model_dir, **model_kwargs)

    audio_duration = pydub.AudioSegment.from_file(args.audio).duration_seconds
    transcribe_fn = transcribe_long if audio_duration > SPEECH_LENGTH else transcribe
    transcribe_params = {
        "model": model_instance,
        "audio": args.audio,
        "lang_sym": args.lang_sym,
        "region_sym": args.region_sym,
        "predict_time": args.predict_time,
        "padding_speech": args.padding_speech,
    }
    transcribe_fn(**transcribe_params)


if __name__ == "__main__":
    cli()



# encoding: utf8

import logging
import numpy as np
import subprocess

from .constants import SAMPLE_RATE


logger = logging.getLogger("dolphin")


# copy from whisper
def load_audio(file: str, sr: int = SAMPLE_RATE):
    """
    Open an audio file and read as mono waveform, resampling as necessary.
    """

    cmd = [
        "ffmpeg",
        "-nostdin",
        "-threads", "0",
        "-i", file,
        "-f", "s16le",
        "-ac", "1",
        "-acodec", "pcm_s16le",
        "-ar", str(sr),
        "-"
    ]

    try:
        out = subprocess.run(cmd, capture_output=True, check=True).stdout
    except Exception as e:
        logger.error(f"load audio error, {e}")
        raise RuntimeError(f"Failed to load audio: {e.stderr.decode()}") from e

    return np.frombuffer(out, np.int16).flatten().astype(np.float32) / 32768.0


def convert_audio(inp: str, outp: str):
    try:
        cmd = ['ffmpeg',
            '-loglevel', '0',
            '-nostdin',
            '-threads', '0',
            '-i', inp,
            '-acodec', 'pcm_s16le',
            '-ac', '1',
            '-ar', '16000',
            outp,
            '-y'
        ]
        subprocess.run(cmd, capture_output=False, check=True)
    except Exception as e:
        logger.error(f"convert audio error, {e}")
        raise RuntimeError(f"Failed to convert audio: {e.stderr.decode()}") from e



# encoding: utf8

import os
import time
import yaml
import logging
import dataclasses
import argparse
from pathlib import Path
from os.path import join, dirname, abspath
from typing import Union, Optional, List, Tuple, Dict

import numpy as np
import torch
import torch.nn.functional as F
from typeguard import typechecked
from espnet2.tasks.lm import LMTask
from espnet2.tasks.s2t import S2TTask
from espnet2.torch_utils.device_funcs import to_device
from espnet.nets.batch_beam_search import BatchBeamSearch
from espnet2.train.abs_espnet_model import AbsESPnetModel
from espnet.nets.scorer_interface import BatchScorerInterface
from espnet.nets.scorers.length_bonus import LengthBonus
from espnet2.text.build_tokenizer import build_tokenizer
from espnet2.text.token_id_converter import TokenIDConverter
from espnet2.bin.s2t_inference import Speech2Text, ListOfHypothesis
from dolphin.scorefilter import DolphinScoreFilter
from dolphin.beam_search import SimpleBeamSearch
from espnet.nets.beam_search import BeamSearch, Hypothesis

from .constants import (SPEECH_LENGTH, SAMPLE_RATE, FIRST_TIME_SYMBOL, LAST_TIME_SYMBOL, NOTIME_SYMBOL,
                        FIRST_LANG_SYMBOL, LAST_LANG_SYMBOL, FIRST_REGION_SYMBOL, LAST_REGION_SYMBOL)


logger = logging.getLogger("dolphin")


@dataclasses.dataclass
class TranscribeResult:
    text: str
    text_nospecial: str
    language: str
    region: str
    rtf: float


@dataclasses.dataclass
class TranscribeSegmentResult:
    text: str
    text_nospecial: str
    language: str
    region: str
    rtf: float
    start: float
    end: float


class DolphinSpeech2Text(Speech2Text):

    @typechecked
    def __init__(
        self,
        s2t_train_config: Union[Path, str, Dict] = None,
        s2t_model_file: Union[Path, str, None] = None,
        lm_train_config: Union[Path, str, None] = None,
        lm_file: Union[Path, str, None] = None,
        ngram_scorer: str = "full",
        ngram_file: Union[Path, str, None] = None,
        device: str = "cpu",
        maxlenratio: float = 0.0,
        minlenratio: float = 0.0,
        batch_size: int = 1,
        dtype: str = "float32",
        beam_size: int = 5,
        lm_weight: float = 0.0,
        ngram_weight: float = 0.0,
        penalty: float = 0.0,
        nbest: int = 1,
        normalize_length: bool = False,
        quantize_s2t_model: bool = False,
        quantize_lm: bool = False,
        quantize_modules: List[str] = ["Linear"],
        quantize_dtype: str = "qint8",
        task_sym: str = "<asr>",
        predict_time: bool = True,
        **kwargs,
    ):

        qconfig_spec = set([getattr(torch.nn, q) for q in quantize_modules])
        quantize_dtype: torch.dtype = getattr(torch, quantize_dtype)

        # 1. Build S2T model
        s2t_model, s2t_train_args = self.build_model_from_file(s2t_train_config, s2t_model_file, device)
        s2t_model.to(dtype=getattr(torch, dtype)).eval()

        if quantize_s2t_model:
            logger.info("Use quantized s2t model for decoding.")
            s2t_model = torch.quantization.quantize_dynamic(s2t_model, qconfig_spec=qconfig_spec, dtype=quantize_dtype)

        decoder = s2t_model.decoder
        token_list = s2t_model.token_list
        scorers = dict(
            decoder=decoder,
            length_bonus=LengthBonus(len(token_list)),
            scorefilter=DolphinScoreFilter(
                notimestamps=token_list.index(NOTIME_SYMBOL),
                first_time=token_list.index(FIRST_TIME_SYMBOL),
                last_time=token_list.index(LAST_TIME_SYMBOL),
                sos=s2t_model.sos,
                eos=s2t_model.eos,
                vocab_size=len(token_list),
            ),
        )

        # 2. Build language model
        if lm_train_config is not None:
            lm, lm_train_args = LMTask.build_model_from_file(lm_train_config, lm_file, device)

            if quantize_lm:
                logger.info("Use quantized lm for decoding.")
                lm = torch.quantization.quantize_dynamic(lm, qconfig_spec=qconfig_spec, dtype=quantize_dtype)

            scorers["lm"] = lm.lm

        # 3. Build ngram model
        if ngram_file is not None:
            if ngram_scorer == "full":
                from espnet.nets.scorers.ngram import NgramFullScorer

                ngram = NgramFullScorer(ngram_file, token_list)
            else:
                from espnet.nets.scorers.ngram import NgramPartScorer

                ngram = NgramPartScorer(ngram_file, token_list)
            scorers["ngram"] = ngram

        # 4. Build BeamSearch object
        weights = dict(
            decoder=1.0,
            lm=lm_weight,
            ngram=ngram_weight,
            length_bonus=penalty,
            scorefilter=1.0,
        )
            
        beam_search = SimpleBeamSearch(
            beam_size=beam_size,
            weights=weights,
            scorers=scorers,
            sos=s2t_model.sos,
            eos=s2t_model.eos,
            vocab_size=len(token_list),
            token_list=token_list,
            pre_beam_score_key="full",
            normalize_length=normalize_length,
        )

        if batch_size == 1:
            non_batch = [
                k
                for k, v in beam_search.full_scorers.items()
                if not isinstance(v, BatchScorerInterface)
            ]
            if len(non_batch) == 0:
                beam_search.__class__ = SimpleBeamSearch
                logger.info("BatchBeamSearch implementation is selected.")
            else:
                logger.warning(
                    f"As non-batch scorers {non_batch} are found, "
                    f"fall back to non-batch implementation."
                )

            beam_search.to(device=device, dtype=getattr(torch, dtype)).eval()
            for scorer in scorers.values():
                if isinstance(scorer, torch.nn.Module):
                    scorer.to(device=device, dtype=getattr(torch, dtype)).eval()
            logger.info(f"Decoding device={device}, dtype={dtype}")

        bpemodel = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets/bpe.model")
        tokenizer = build_tokenizer(token_type="bpe", bpemodel=bpemodel)
        converter = TokenIDConverter(token_list=token_list)
        logger.info(f"Text tokenizer: {tokenizer}")

        self.s2t_model = s2t_model
        self.s2t_train_args = s2t_train_args
        self.preprocessor_conf = s2t_train_args.preprocessor_conf
        self.converter = converter
        self.tokenizer = tokenizer
        self.beam_search = beam_search
        self.maxlenratio = maxlenratio
        self.minlenratio = minlenratio
        self.device = device
        self.dtype = dtype
        self.nbest = nbest

        self.task_sym = task_sym
        self.predict_time = predict_time

    @staticmethod
    @typechecked
    def build_model_from_file(
        config: Optional[Union[Path, str, Dict]] = None,
        model_file: Optional[Union[Path, str]] = None,
        device: str = "cpu",
    ) -> Tuple[AbsESPnetModel, argparse.Namespace]:
        """Build model from the files.

        This method is used for inference or fine-tuning.

        Args:
            config_file: The yaml file saved when training.
            model_file: The model file saved when training.
            device: Device type, "cpu", "cuda", or "cuda:N".

        """
        if isinstance(config, dict):
            args = argparse.Namespace(**config)
        else:
            with open(config, "r", encoding="utf-8") as f:
                args = yaml.safe_load(f)
            args = argparse.Namespace(**args)
        args.normalize_conf["stats_file"] = os.path.join(dirname(abspath(__file__)), "assets/feats_stats.npz")

        model = S2TTask.build_model(args)
        if not isinstance(model, AbsESPnetModel):
            raise RuntimeError(f"model must inherit {AbsESPnetModel.__name__}, but got {type(model)}")

        if device == "mps":
            model.to(device, dtype=torch.float32)
        else:
            model.to(device)

        if model_file is not None:
            if device == "cuda":
                device = f"cuda:{torch.cuda.current_device()}"

            state_dict = torch.load(model_file, map_location=device)
            model.load_state_dict(state_dict, strict=False)

        return model, args

    def detect_language(self, speech: torch.Tensor, lang_sym: str = None, with_enc_output: bool = False, **kwargs) -> Tuple[int, int]:
        """
        Detect language and region.

        Args:
            speech: waveform

        Returns:
            tuple(lang_id, region_id)
        """
        # Preapre speech
        if isinstance(speech, np.ndarray):
            speech = torch.tensor(speech)

        # Only support single-channel speech
        if speech.dim() > 1:
            assert (speech.dim() == 2 and speech.size(1) == 1), f"speech of size {speech.size()} is not supported"
            speech = speech.squeeze(1)  # (nsamples, 1) --> (nsamples,)

        speech_length = int(SAMPLE_RATE * SPEECH_LENGTH)
        # Pad or trim speech to the fixed length
        if speech.size(-1) >= speech_length:
            speech = speech[:speech_length]

        # Batchify input
        # speech: (nsamples,) -> (1, nsamples)
        speech = speech.unsqueeze(0).to(getattr(torch, self.dtype))
        # lengths: (1,)
        lengths = speech.new_full([1], dtype=torch.long, fill_value=speech.size(1))
        batch = {"speech": speech, "speech_lengths": lengths}

        batch = to_device(batch, device=self.device)
        enc, enc_olens = self.s2t_model.encode(**batch)
        assert enc.size(0) == 1

        lang_first_id = self.converter.token2id[FIRST_LANG_SYMBOL]
        lang_last_id = self.converter.token2id[LAST_LANG_SYMBOL]
        region_first_id = self.converter.token2id[FIRST_REGION_SYMBOL]
        region_last_id = self.converter.token2id[LAST_REGION_SYMBOL]

        decoder = self.s2t_model.decoder
        # detect language
        lang_id = None
        if lang_sym is None:
            ys = torch.tensor([self.s2t_model.sos], dtype=torch.long, device=self.device).unsqueeze(0)
            logp, _ = decoder.batch_score(ys, [None], enc)
            mask = torch.ones(logp.size(1), dtype=torch.bool)
            mask[lang_first_id:lang_last_id+1] = False
            logp[0, mask] = -np.inf
            lang_id = logp.argmax(dim=-1).tolist()[0]
        else:
            logger.debug(f"lang symbol is given: {lang_sym}")
            lang_id = self.converter.token2id[f"<{lang_sym}>"]

        # detect region
        ys = torch.tensor([self.s2t_model.sos, lang_id], dtype=torch.long, device=self.device).unsqueeze(0)
        logp, _ = decoder.batch_score(ys, [None], enc)
        mask = torch.ones(logp.size(1), dtype=torch.bool)
        mask[region_first_id:region_last_id+1] = False
        logp[0, mask] = -np.inf
        region_id = logp.argmax(dim=-1).tolist()[0]

        lang_symbol = self.converter.ids2tokens([lang_id])[0]
        region_symbol = self.converter.ids2tokens([region_id])[0]

        logger.debug(f"detect language: {lang_symbol}, region: {region_symbol}")

        if with_enc_output:
            return lang_id, region_id, enc
        else:
            return lang_id, region_id, None

    @torch.no_grad()
    @typechecked
    def __call__(
        self,
        speech: Union[torch.Tensor, np.ndarray],
        lang_sym: Optional[str] = None,
        region_sym: Optional[str] = None,
        predict_time: Optional[bool] = None,
        padding_speech: bool = False,
    ) -> TranscribeResult:
        """Inference for a single utterance.

        The input speech will be padded or trimmed to the fixed length,
        which is consistent with training.

        Args:
            speech: input speech of shape (nsamples,) or (nsamples, nchannels=1)
            lang_sym: language code symbol (e.g. zh)
            region_sym: region symbol (e.g. CN)
            predict_time: whether to predict timestamps
            padding_speech: whether to padding speech to 30 seconds, default is true

        Returns:
            TranscribeResult

        """

        predict_time = predict_time if predict_time is not None else self.predict_time

        start_tm = time.monotonic()
        enc = None
        if all([lang_sym, region_sym]):
            lang_id = self.converter.token2id[f"<{lang_sym}>"]
            region_id = self.converter.token2id[f"<{region_sym}>"]
        else:

            lang_id, region_id, enc = self.detect_language(speech, lang_sym, with_enc_output=True)

        task_id = self.converter.token2id["<asr>"]
        notime_id = self.converter.token2id[NOTIME_SYMBOL]

        # Prepare hyp_primer
        hyp_primer = [self.s2t_model.sos, lang_id, region_id, task_id]
        if not predict_time:
            hyp_primer.append(notime_id)

        self.beam_search.set_hyp_primer(hyp_primer)

        # Preapre speech
        if isinstance(speech, np.ndarray):
            speech = torch.tensor(speech)

        # Only support single-channel speech
        if speech.dim() > 1:
            assert (speech.dim() == 2 and speech.size(1) == 1), f"speech of size {speech.size()} is not supported"
            speech = speech.squeeze(1)  # (nsamples, 1) --> (nsamples,)

        nsamples = speech.size(-1)
        speech_length = int(SAMPLE_RATE * SPEECH_LENGTH)
        # Pad or trim speech to the fixed length
        if speech.size(-1) >= speech_length:
            speech = speech[:speech_length]
        else:
            speech = F.pad(speech, (0, speech_length - speech.size(-1))) if padding_speech else speech

        # Batchify input
        # speech: (nsamples,) -> (1, nsamples)
        speech = speech.unsqueeze(0).to(getattr(torch, self.dtype))
        # lengths: (1,)
        lengths = speech.new_full([1], dtype=torch.long, fill_value=speech.size(1))
        batch = {"speech": speech, "speech_lengths": lengths}
        logger.debug("speech length: " + str(speech.size(1)))

        # a. To device
        batch = to_device(batch, device=self.device)

        # b. Forward Encoder
        if enc is None:
            enc, enc_olens = self.s2t_model.encode(**batch)
        if isinstance(enc, tuple):
            enc, _ = enc
        encode_tm = time.monotonic()
        logger.debug(f"encode time: {round(encode_tm - start_tm, 2)} seconds")

        assert len(enc) == 1, len(enc)

        # c. Pass the encoder result to the beam search
        results = self._decode_single_sample(enc[0])
        text, _, _, text_nospecial, _ = results[0]
        decode_tm = time.monotonic()
        logger.debug(f"decode time: {round(decode_tm - encode_tm, 2)} senconds")
        rtf = round((decode_tm - start_tm) / (nsamples / SAMPLE_RATE), 2)

        lang, region = self.converter.ids2tokens([lang_id, region_id])
        ret = TranscribeResult(text, text_nospecial, lang[1:-1], region[1:-1], rtf=rtf)
        return ret



# encoding: utf8

LANGUAGE_CODES = {
    "zh": ("Mandarin Chinese", "中文"),
    "ja": ("Japanese", "日语"),
    "th": ("Thai", "泰语"),
    "ru": ("Russian", "俄语"),
    "ko": ("Korean", "韩语"),
    "id": ("Indonesian", "印度尼西亚语"),
    "vi": ("Vietnamese", "越南语"),
    "ct": ("Yue Chinese", "粤语"),
    "hi": ("Hindi", "印地语"),
    "ur": ("Urdu", "乌尔都语"),
    "ms": ("Malay", "马来语"),
    "uz": ("Uzbek", "乌兹别克语"),
    "ar": ("Arabic", "阿拉伯语"),
    "fa": ("Persian", "波斯语"),
    "bn": ("Bengali", "孟加拉语"),
    "ta": ("Tamil", "泰米尔语"),
    "te": ("Telugu", "泰卢固语"),
    "ug": ("Uighur", "维吾尔语"),
    "gu": ("Gujarati", "古吉拉特语"),
    "my": ("Burmese", "缅甸语"),
    "tl": ("Tagalog", "塔加洛语"),
    "kk": ("Kazakh", "哈萨克语"),
    "or": ("Oriya / Odia", "奥里亚语"),
    "ne": ("Nepali", "尼泊尔语"),
    "mn": ("Mongolian", "蒙古语"),
    "km": ("Khmer", "高棉语"),
    "jv": ("Javanese", "爪哇语"),
    "lo": ("Lao", "老挝语"),
    "si": ("Sinhala", "僧伽罗语"),
    "fil": ("Filipino", "菲律宾语"),
    "ps": ("Pushto", "普什图语"),
    "pa": ("Panjabi", "旁遮普语"),
    "kab": ("Kabyle", "卡拜尔语"),
    "ba": ("Bashkir", "巴什基尔语"),
    "ks": ("Kashmiri", "克什米尔语"),
    "tg": ("Tajik", "塔吉克语"),
    "su": ("Sundanese", "巽他语"),
    "mr": ("Marathi", "马拉地语"),
    "ky": ("Kirghiz", "吉尔吉斯语"),
    "az": ("Azerbaijani", "阿塞拜疆语")
}

LANGUAGE_REGION_CODES = {
    "zh-CN": ("Chinese (Mandarin)", "中文(普通话)"),
    "zh-TW": ("Chinese (Taiwan)", "中文(台湾)"),
    "zh-WU": ("Chinese (Wuyu)", "中文(吴语)"),
    "zh-SICHUAN": ("Chinese (Sichuan)", "中文(四川话)"),
    "zh-SHANXI": ("Chinese (Shanxi)", "中文(山西话)"),
    "zh-ANHUI": ("Chinese (Anhui)", "中文(安徽话)"),
    "zh-TIANJIN": ("Chinese (Tianjin)", "中文(天津话)"),
    "zh-NINGXIA": ("Chinese (Ningxia)", "中文(宁夏话)"),
    "zh-SHAANXI": ("Chinese (Shanxi)", "中文(陕西话)"),
    "zh-HEBEI": ("Chinese (Hebei)", "中文(河北话)"),
    "zh-SHANDONG": ("Chinese (Shandong)", "中文(山东话)"),
    "zh-GUANGDONG": ("Chinese (Guangdong)", "中文(广东话)"),
    "zh-SHANGHAI": ("Chinese (Shanghai)", "中文(上海话)"),
    "zh-HUBEI": ("Chinese (Hubei)", "中文(湖北话)"),
    "zh-LIAONING": ("Chinese (Liaoning)", "中文(辽宁话)"),
    "zh-GANSU": ("Chinese (Gansu)", "中文(甘肃话)"),
    "zh-FUJIAN": ("Chinese (Fujian)", "中文(福建话)"),
    "zh-HUNAN": ("Chinese (Hunan)", "中文(湖南话)"),
    "zh-HENAN": ("Chinese (Henan)", "中文(河南话)"),
    "zh-YUNNAN": ("Chinese (Yunnan)", "中文(云南话)"),
    "zh-MINNAN": ("Chinese (Minnan)", "中文(闽南语)"),
    "zh-WENZHOU": ("Chinese (Wenzhou)", "中文(温州话)"),
    "ja-JP": ("Japanese", "日语"),
    "th-TH": ("Thai", "泰语"),
    "ru-RU": ("Russian", "俄语"),
    "ko-KR": ("Korean", "韩语"),
    "id-ID": ("Indonesian", "印度尼西亚语"),
    "vi-VN": ("Vietnamese", "越南语"),
    "ct-NULL": ("Yue (Chinese)", "粤语"),
    "ct-HK": ("Yue (Hongkong)", "粤语(香港)"),
    "ct-GZ": ("Yue (Guangdong)", "粤语(广东)"),
    "hi-IN": ("Hindi", "印地语"),
    "ur-IN": ("Urdu", "乌尔都语(印度)"),
    "ur-PK": ("Urdu (Islamic Republic of Pakistan)", "乌尔都语"),
    "ms-MY": ("Malay", "马来语"),
    "uz-UZ": ("Uzbek", "乌兹别克语"),
    "ar-MA": ("Arabic (Morocco)", "阿拉伯语(摩洛哥)"),
    "ar-GLA": ("Arabic", "阿拉伯语"),
    "ar-SA": ("Arabic (Saudi Arabia)", "阿拉伯语(沙特)"),
    "ar-EG": ("Arabic (Egypt)", "阿拉伯语(埃及)"),
    "ar-KW": ("Arabic (Kuwait)", "阿拉伯语(科威特)"),
    "ar-LY": ("Arabic (Libya)", "阿拉伯语(利比亚)"),
    "ar-JO": ("Arabic (Jordan)", "阿拉伯语(约旦)"),
    "ar-AE": ("Arabic (U.A.E.)", "阿拉伯语(阿联酋)"),
    "ar-LVT": ("Arabic (Levant)", "阿拉伯语(黎凡特)"),
    "fa-IR": ("Persian", "波斯语"),
    "bn-BD": ("Bengali", "孟加拉语"),
    "ta-SG": ("Tamil (Singaporean)", "泰米尔语(新加坡)"),
    "ta-LK": ("Tamil (Sri Lankan)", "泰米尔语(斯里兰卡)"),
    "ta-IN": ("Tamil (India)", "泰米尔语(印度)"),
    "ta-MY": ("Tamil (Malaysia)", "泰米尔语(马来西亚)"),
    "te-IN": ("Telugu", "泰卢固语"),
    "ug-NULL": ("Uighur", "维吾尔语"),
    "ug-CN": ("Uighur", "维吾尔语"),
    "gu-IN": ("Gujarati", "古吉拉特语"),
    "my-MM": ("Burmese", "缅甸语"),
    "tl-PH": ("Tagalog", "塔加洛语"),
    "kk-KZ": ("Kazakh", "哈萨克语"),
    "or-IN": ("Oriya / Odia", "奥里亚语"),
    "ne-NP": ("Nepali", "尼泊尔语"),
    "mn-MN": ("Mongolian", "蒙古语"),
    "km-KH": ("Khmer", "高棉语"),
    "jv-ID": ("Javanese", "爪哇语"),
    "lo-LA": ("Lao", "老挝语"),
    "si-LK": ("Sinhala", "僧伽罗语"),
    "fil-PH": ("Filipino", "菲律宾语"),
    "ps-AF": ("Pushto", "普什图语"),
    "pa-IN": ("Panjabi", "旁遮普语"),
    "kab-NULL": ("Kabyle", "卡拜尔语"),
    "ba-NULL": ("Bashkir", "巴什基尔语"),
    "ks-IN": ("Kashmiri", "克什米尔语"),
    "tg-TJ": ("Tajik", "塔吉克语"),
    "su-ID": ("Sundanese", "巽他语"),
    "mr-IN": ("Marathi", "马拉地语"),
    "ky-KG": ("Kirghiz", "吉尔吉斯语"),
    "az-AZ": ("Azerbaijani", "阿塞拜疆语"),
}



import os
import pkg_resources
from setuptools import find_packages, setup


def read_version(fname="dolphin/version.py"):
    exec(compile(open(fname, encoding="utf-8").read(), fname, "exec"))
    return locals()["__version__"]


requirements = []
setup(
    name="dataoceanai-dolphin",
    py_modules=["dolphin"],
    version=read_version(),
    description="DataoceanAI Open-source Large Speech Model",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    readme="README.md",
    python_requires=">=3.7",
    author="DataoceanAI",
    url="https://github.com/DataoceanAI/dolphin",
    license="Apache-2.0",
    license_files="LICENSE",
    packages=[
        "dolphin",
        "dolphin/assets",
    ],
    package_data={
        "dolphin/assets": ["*"]
    },
    install_requires=requirements
    + [
        str(r)
        for r in pkg_resources.parse_requirements(
            open(os.path.join(os.path.dirname(__file__), "requirements.txt"))
        )
    ],
    entry_points={
        "console_scripts": ["dolphin=dolphin.transcribe:cli"],
    },
    include_package_data=True,
    extras_require={"dev": ["pytest", "flake8"]},
)